/*
 * File: lookout.js
 *   LookOut Mozilla MailNews Attachments Javascript Overlay for TNEF
 *
 * Copyright:
 *   Copyright (C) 2007 Aron Rubin <arubin@atl.lmco.com>
 *
 * About:
 *   Benevolently hijack the Mozilla mailnews attachment list and expand all
 *   Transport Neutral Encapsulation Format (TNEF) encoded attchments. When
 *   TNEF attachments are opened, decode them.
 */

/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is LookOut.
 *
 * The Initial Developer of the Original Code is
 * Aron Rubin.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


// Since we're automatically downloading, we don't get the file picker's
// logic to check for existing files, so we need to do that here.
//
// Note - this code is identical to that in contentAreaUtils.js.
// If you are updating this code, update that code too! We can't share code
// here since this is called in a js component.
function find_unique_filename( aLocalFile ) {
  var uniqifier_re = /(-\d+)?(\.[^.]+)?$/;
  var parts = uniqifier_re.exec( aLocalFile.leafName );
  var prefix = "";
  var uniqifier = 0;
  var postfix = "";

  if( parts && parts.index >= 0 ) {
    debugDump( aLocalFile.path + " -> " + parts.toSource(), 7 );
    prefix = aLocalFile.leafName.slice( 0, parts.index - 1 );
    if( parts[1] != undefined )
      uniqifier = parseInt( parts[1].substr( 1 ) ); // chop '-'
    if( parts[2] != undefined )
      postfix = parts[2];
  } else {
    prefix = aLocalFile.leafName;
  }

  while( aLocalFile.exists() ) {
    uniqifier++;
    aLocalFile.leafName = prefix + "-" + uniqifier + postfix;
  }
  return( aLocalFile );
}

function make_temp_file( filename ) {
  var file_locator = Components.classes["@mozilla.org/file/directory_service;1"]
                     .getService( Components.interfaces.nsIProperties );
  var temp_dir = file_locator.get( "TmpD", Components.interfaces.nsIFile );

  var local_target = temp_dir.clone();
  local_target.append( filename );

  return( find_unique_filename( local_target ) );
}

function LookoutStreamListener() {
}
LookoutStreamListener.prototype = {
  attachment: null,
  mAttUrl: null,
  mMsgUri: null,
  mStream: null,
  mPackage: null,
  mPartId: 1,
  do_open: false,
  req_part_id: 0,

  stream_started: false,
  cur_outfilestrm: null,
  cur_outstrm: null,
  cur_content_type: null,
  cur_length: 0,
  cur_date: null,
  cur_url: "",

  QueryInterface: function ( iid )  {
    if( iid.equals( Components.interfaces.nsIStreamListener ) ||
        iid.equals( Components.interfaces.nsISupports ) )
      return this;

    throw Components.results.NS_NOINTERFACE;
    return( 0 );
  },

  onStartRequest: function ( aRequest, aContext ) {
    this.mStream = Components.classes['@mozilla.org/binaryinputstream;1'].
                    createInstance( Components.interfaces.nsIBinaryInputStream );
  },
    
  onStopRequest: function ( aRequest, aContext, aStatusCode ) {
    var channel = aRequest.QueryInterface(Components.interfaces.nsIChannel);
    var fsm;
    try {
      fsm = GetDBView().URIForFirstSelectedMessage;
    } catch(ex) {
      fsm = this.mMsgUri; // continue in single message view
    }

    if( !(this.mMsgUri == fsm && this.mStream) ) {
      debugDump( "LookOut: strange things a foot", 5 );
      aRequest.cancel( Components.results.NS_BINDING_ABORTED );
      return;
    }

    this.mPartId++;
    this.mStream = null;
    this.stream_started = false;
    this.mPackage = null;
  },

  onDataAvailable: function ( aRequest, aContext, aInputStream, aOffset, aCount ) {
    var fsm;
    try {
      fsm = GetDBView().URIForFirstSelectedMessage;
    } catch(ex) {
      fsm = this.mMsgUri; // continue in single message view
    }

    if( this.mMsgUri != fsm ) {
      debugDump( "LookOut: data available wrong", 5 );
      aRequest.cancel( Components.results.NS_BINDING_ABORTED );
      return;
    }
    if( !this.stream_started ) {
      this.mStream.setInputStream( aInputStream );
      this.stream_started = true;
    }

    this.mPackage = tnef_pack_parse_stream( this.mStream, this, this.mPackage );
  },

  onTnefStart: function ( filename, content_type, length, date ) {
    if( !content_type )
      content_type = "application/binary";

    if( this.do_open ) {
      if( !this.req_part_id || this.mPartId == this.req_part_id ) {
        var outfile = make_temp_file( filename );
        var ios = Components.classes["@mozilla.org/network/io-service;1"]
                  .getService(Components.interfaces.nsIIOService);
        var url = ios.newFileURI( outfile ); // only needed, when 'do_open'

        this.cur_outfilestrm = Components.classes["@mozilla.org/network/file-output-stream;1"]
                               .createInstance( Components.interfaces.nsIFileOutputStream );
        this.cur_outfilestrm.init( outfile, 0x02 | 0x08, 0666, 0 );

        this.cur_outstrm = Components.classes["@mozilla.org/binaryoutputstream;1"]
                           .createInstance( Components.interfaces.nsIBinaryOutputStream );
        this.cur_outstrm.setOutputStream( this.cur_outfilestrm );

        this.cur_filename = filename;
        this.cur_content_type = content_type;
        this.cur_length = length;
        this.cur_date = date;
        this.cur_url = url;
      }
    } else {
      var mimeurl = this.mAttUrl + "." + this.mPartId;
      dump( "adding attachment: " + mimeurl );
      lookout_lib.add_sub_attachment_to_list( this.attachment, content_type, filename,
                                              this.mPartId, mimeurl, this.mMsgUri, length );
    }
    debugDump( "LookOut: onTnefStart\nParent: "+this.attachment
              +"\nmMsgUri: "+this.mMsgUri
              +"\nrequested Part_ID: "+this.req_part_id
              +"\nPart_ID: "+this.mPartId
              +"\nDisplayname: "+filename.split("\0")[0]
              +"\nContent-Type: "+content_type.split("\0")[0]
              +"\nLength: "+length
              +"\nURL: "+(url ? url.spec : "")
              +"\nmimeurl: "+(mimeurl ? mimeurl : ""), 7 );
  },

  onTnefEnd: function ( ) {
    debugDump( "LookOut: onTnefEnd", 8 );
    if( this.cur_outfilestrm )
      this.cur_outfilestrm.close();

    if( this.do_open && (!this.req_part_id || this.mPartId == this.req_part_id) ) {
      debugDump( "Opening attachment '"+this.cur_url.spec+"'", 7 );
      messenger.OpenURL( this.cur_url.spec );
    }

    this.cur_outfilestrm = null;
    this.cur_outstrm = null;
    this.cur_content_type = null;
    this.cur_length = 0;
    this.cur_date = null;
    this.mPartId++;
  },

  onTnefData: function ( position, data ) {
    debugDump( "LookOut: onTnefData position" + position + "  data " + data.length + "  outstrm " + this.cur_outstrm, 8 );
    if( this.cur_outstrm ) {
      if( data ) {
        debugDump( "LookOut: writing " + data.length + "bytes to file", 7 );
        this.cur_outstrm.write( data, data.length );
      }
    }
  }
}  

var lookout_lib = {
  orig_openAttachment: null,
  orig_saveAttachment: null,

  onload: function() {
    var listener = {};
    listener.onStartHeaders = lookout_lib.on_start_headers;
    listener.onEndHeaders = lookout_lib.on_end_headers;
    gMessageListeners.push( listener );

    // FIXME - fix mozilla so there is a cleaner way here
    // benevolently hijack the openAttachment and saveAttachment functions
    if( openAttachment ) {
      lookout_lib.orig_openAttachment = openAttachment;
      openAttachment = lookout_lib.open_attachment;
    }
    if( saveAttachment ) {
      lookout_lib.orig_saveAttachment = saveAttachment;
      saveAttachment = lookout_lib.save_attachment;
    }
  },

  scan_for_tnef: function ( ) {
    var messenger2 = Components.classes["@mozilla.org/messenger;1"]
                    .getService(Components.interfaces.nsIMessenger);

    // for each attachment of the current message
    for( index in currentAttachments ) {
      var attachment = currentAttachments[index];
      debugDump( attachment.toSource(), 8 );
      // we only decode tnef files
      if( (/^application\/ms-tnef/i).test( attachment.contentType ) ) {
        debugDump( "LookOut: found tnef", 7 );

        // open the attachment and look inside
        var stream_listener = new LookoutStreamListener(); 
        stream_listener.attachment = attachment;
        stream_listener.mAttUrl = attachment.url;
        if( attachment.uri )
          stream_listener.mMsgUri = attachment.uri;
        else
          stream_listener.mMsgUri = attachment.messageUri;
        stream_listener.do_open = false;

        var mms = messenger2.messageServiceFromURI( stream_listener.mMsgUri )
                   .QueryInterface( Components.interfaces.nsIMsgMessageService );
        mms.openAttachment( attachment.contentType, attachment.displayName,
                            attachment.url, stream_listener.mMsgUri, stream_listener, 
                            null, null );
      }
    }
  },

  add_sub_attachment_to_list: function ( parent, content_type, display_name, part_id, atturl, msguri, length ) {
    /*
    var attachmentList = document.getElementById( 'attachmentList' );
    var i = 0;

    for( i = 0; i < attachmentList.childNodes.length &&
                attachmentList.childNodes[i].attachment.url != attUrl; i++ );

    // if we found our list item then we are done
    if( i >= attachmentList.childNodes.length )
      return;
    */
    debugDump( " content_type:"+content_type +", atturl:"+ atturl +",  display_name:"+ display_name +", msguri:"+ msguri, 8 );
    var attachment = new createNewAttachmentInfo( content_type, atturl, display_name, msguri, true )
    attachment.parent = parent;
    attachment.part_id = part_id;
    currentAttachments.push( attachment );
    debugDump( attachment.toSource(), 8 );
    var attachmentList = document.getElementById('attachmentList');
    var item = attachmentList.appendItem( attachment.displayName, "" );
    item.setAttribute( "class", "descriptionitem-iconic" ); 
    item.setAttribute( "tooltip", "attachmentListTooltip" );
    item.attachment = attachment;
    item.setAttribute( "attachmentUrl", attachment.url );
    item.setAttribute( "attachmentContentType", attachment.contentType );
    item.setAttribute( "attachmentUri", attachment.uri );
    // also set orientation to preserve TB 1.5 compatibility
    // ( copied from http://lxr.mozilla.org/mozilla1.7/source/mail/base/content/msgHdrViewOverlay.js#1121 )
    if (gShowLargeAttachmentView)
    {
      item.setAttribute("largeView", "true");
      item.setAttribute("orient", "vertical");
    }
    setApplicationIconForAttachment( attachment, item, gShowLargeAttachmentView );

    // try to call "Attachment Sizes", extension {90ceaf60-169c-40fb-b224-7204488f061d}
    try {
      ABglobals.setAttSizeTextFor(atturl, length, false);
    } catch(ex) {}
  },

  open_attachment: function ( attachment ) {
    debugDump( attachment.toSource(), 8 );

    if( !attachment.parent ||
        !(/^application\/ms-tnef/i).test( attachment.parent.contentType ) ) {
      lookout_lib.orig_openAttachment( attachment );
      return;
    }

    var messenger2 = Components.classes["@mozilla.org/messenger;1"]
                    .getService(Components.interfaces.nsIMessenger);
    var stream_listener = new LookoutStreamListener();
    stream_listener.req_part_id = attachment.part_id;
    stream_listener.mAttUrl = attachment.parent.url;
    if( attachment.uri )
      stream_listener.mMsgUri = attachment.uri;
    else
      stream_listener.mMsgUri = attachment.messageUri;
    stream_listener.do_open = true;
    debugDump( "open_attachment\nParent: "+(attachment.parent == null ? "-" : attachment.parent.url)
              +"\nContent-Type: "+attachment.contentType.split("\0")[0]
              +"\nDisplayname: "+attachment.displayName.split("\0")[0]
              +"\nPart_ID: "+attachment.part_id
              +"\nisExternal: "+attachment.isExternalAttachment
              +"\nURL: "+attachment.url
              +"\nmMsgUri: "+stream_listener.mMsgUri, 7 );
    var mms = messenger2.messageServiceFromURI( stream_listener.mMsgUri )
              .QueryInterface( Components.interfaces.nsIMsgMessageService );
    mms.openAttachment( attachment.parent.contentType, attachment.parent.displayName,
                        attachment.parent.url, stream_listener.mMsgUri, stream_listener, 
                        null, null );
  },

  save_attachment: function ( attachment ) {
    if( !attachment.parent ||
        !(/^application\/ms-tnef/i).test( attachment.parent.contentType ) ) {
      lookout_lib.orig_saveAttachment( attachment );
      return;
    }

    var messenger2 = Components.classes["@mozilla.org/messenger;1"]
                    .getService(Components.interfaces.nsIMessenger);
    var stream_listener = new LookoutStreamListener(); 
    stream_listener.req_part_id = attachment.part_id;
    stream_listener.mAttUrl = attachment.parent.url;
    if( attachment.uri )
      stream_listener.mMsgUri = attachment.uri;
    else
      stream_listener.mMsgUri = attachment.messageUri;
    stream_listener.do_open = true;
    debugDump( "open_attachment\nParent: "+(attachment.parent == null ? "-" : attachment.parent.url)
              +"\nContent-Type: "+attachment.contentType.split("\0")[0]
              +"\nDisplayname: "+attachment.displayName.split("\0")[0]
              +"\nPart_ID: "+attachment.part_id
              +"\nisExternal: "+attachment.isExternalAttachment
              +"\nURL: "+attachment.url
              +"\nmMsgUri: "+stream_listener.mMsgUri, 7 );
    var mms = messenger2.messageServiceFromURI( stream_listener.mMsgUri )
              .QueryInterface( Components.interfaces.nsIMsgMessageService );
    mms.openAttachment( attachment.parent.contentType, attachment.parent.displayName,
                        attachment.parent.url, stream_listener.mMsgUri, stream_listener, 
                        null, null );
  },

  on_start_headers: function () {
  },

  on_end_headers: function () {
    // defer the call so it is called after all the header work is done
    //   (should have nothing to do with amount of delay)
    setTimeout( lookout_lib.scan_for_tnef, 100 );
  }
}
window.addEventListener( 'load', lookout_lib.onload, false );

